import hashlib
import logging
import requests
from base64 import b64encode
from datetime import datetime
from os import utime
from pathlib import Path

from . import config


class Props:
    def __init__(self, uri=None):
        self.size = None
        self.md5 = None
        self.path = None
        if uri is not None:
            self.path = uri


class LocalFile(Props):
    def __init__(self, f=None):
        super().__init__(f)
        if f:
            self.path = Path(self.path)
            if self.path.is_file():
                self.get_size()
                # self.get_md5()

    def __str__(self):
        return str(self.path)

    def get_size(self):
        if self.path is None:
            return
        self.size = self.path.stat().st_size
        return self.size

    def get_md5(self):
        if self.path is None:
            return
        md5 = hashlib.md5()
        with self.path.open('rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                md5.update(chunk)
        self.md5 = b64encode(md5.digest()).decode('utf-8')
        logging.debug(f"{self.path} MD5: {self.md5}")
        return self.md5

    def get_mtime(self):
        if not self.path:
            return
        return self.path.stat().st_mtime

    def set_mtime(self, http_timestamp):
        if not self.path.is_file():
            logging.error(
                f"Could not set timestamp; non-existant file: {self.path}"
            )
            return
        fmtstr = '%a, %d %b %Y %H:%M:%S %Z'
        timestamp = datetime.strptime(http_timestamp, fmtstr).timestamp()
        utime(self.path, (timestamp, timestamp))


class Url(Props):
    def __init__(
        self,
        url=None,
        request_headers=None,
        timeout=config.HTTP_TIMEOUT
    ):
        super().__init__(url)
        self.request_headers = {}
        if request_headers:
            self.request_headers = request_headers
        self.timeout = timeout
        self.head_response = None
        self.final_url = None
        self.size = None
        self.md5 = None
        self.is_file = None

    def __str__(self):
        return self.path

    def get_head_response(self):
        logging.debug(f"Getting headers from {self.path}.")
        # NOTE: Adding 'identity' encoding to header so that content will be
        # uncompressed, which allows for correct progress measurement. However,
        # this may cause problems with some types of content.
        self.request_headers['Accept-Encoding'] = 'identity'
        try:
            # Force non-compressed txfr
            r = requests.head(
                self.path,
                allow_redirects=True,
                headers=self.request_headers,
                timeout=self.timeout,
            )
        except requests.exceptions.ConnectionError as e:
            if 'Failed to resolve' in str(e):
                logging.error(f"Failed to resolve address: {self.path}")
            else:
                logging.error(f"ConnectionError: {e}")
            return
        except config.HTTP_ERRORS as e:
            logging.error(f"{type(e)}: {e}")
            return
        except Exception as e:
            logging.error(f"{type(e)}: {e}")
            return

        logging.debug(f"head_response headers:{r.headers}")
        self._set_is_file(r.headers)
        self.head_response = r
        self.final_url = r.url
        self.size = self.get_size()
        self.md5 = self.get_md5()
        return self.head_response

    def get_size(self):
        if self.head_response is None:
            r = self.get_head_response()
            if r is None:
                return
        content_length = self.head_response.headers.get('Content-Length')
        content_encoding = self.head_response.headers.get('Content-Encoding')
        if content_encoding is not None:
            logging.critical(f"The server requires receiving the file compressed as '{content_encoding}'.")  # noqa: E501
        logging.debug(f"{content_length=}")
        if content_length:
            self.size = int(content_length)
        return self.size

    def get_md5(self):
        content_md5 = None
        if self.head_response is None:
            r = self.get_head_response()
            if r is None:
                return
        if self.head_response.headers.get('server') == 'AmazonS3':
            # Ref:
            # https://docs.aws.amazon.com/AmazonS3/latest/userguide/checking-object-integrity.html  # noqa: E501
            # https://teppen.io/2018/06/23/aws_s3_etags/
            # i.e. hexmd5(md5(part1)+md5(part2))-{number of parts}
            # So it seems unrealistic to use the etag if it's for a multi-part
            # file. There's a way to generate a local equivalent etag, but that
            # also seems unrealistic:
            # https://teppen.io/2018/10/23/aws_s3_verify_etags/
            if '-' not in self.head_response.headers.get('etag'):
                content_md5 = self.head_response.headers.get('etag')
            if content_md5:
                # Convert from hex to base64
                content_md5_hex = content_md5.strip('"').strip("'")
                content_md5 = b64encode(bytes.fromhex(content_md5_hex)).decode()  # noqa: E501
        else:
            content_md5 = self.head_response.headers.get('Content-MD5')
        if content_md5 is not None:
            content_md5 = content_md5.strip('"').strip("'")
        logging.debug(f"{content_md5=}")
        if content_md5 is not None:
            self.md5 = content_md5
        return self.md5

    def _set_is_file(self, response_headers):
        ''' Determines whether the URL's content is a file or not.
        True: content is downloaded and saved as a local file
        False|None: content is printed to stdout
        '''
        # https://www.w3.org/Protocols/rfc1341/4_Content-Type.html
        content_type = response_headers.get('Content-Type')
        logging.debug(f"{content_type=}")
        content_type_parts = [p.strip() for p in content_type.split(';')]
        mime = content_type_parts[0]
        mime_type, mime_subtype = mime.split('/')
        if mime_type == 'text':
            self.is_file = False
        elif mime_type == 'application' and mime_subtype in ['json', 'xml']:
            self.is_file = False
        else:
            # NOTE: Some types will likely not be handled correctly with a
            # simple "is file or not" property, but hopefully this handles most
            # cases.
            self.is_file = True



import argparse
import logging
import sys
from pathlib import Path
from os import getcwd

from . import config
from .download import Download

__version__ = '0.1.1'


def main():
    parser = argparse.ArgumentParser(prog="net-dl")
    parser.add_argument(
        'url', metavar='URL', type=str, nargs='?',
        help="source URL(s) to download from",
    )
    parser.add_argument(
        '-c', '--continue-download', action='store_true',
        help="attempt to resume a partially-downloaded file"
    )
    parser.add_argument(
        '-d', '--output-directory', type=Path,
        help="destination folder for downloaded file(s)",
    )
    parser.add_argument(
        '-H', '--header', action='append', default=dict(),
        help="add header to the server request (can be repeated): \"X-First-Name: Joe\""  # noqa: E501
    )
    parser.add_argument(
        '-n', '--filename', type=str,
        help="downloaded file's name (overrides name given by server)",
    )
    parser.add_argument(
        '-t', '--timeout', type=int,
        help=f"set server timeout in seconds [default={config.HTTP_TIMEOUT}]",
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help="verbose output",
    )
    parser.add_argument(
        '--version', action='store_true',
        help="show version number",
    )
    parser.add_argument(
        '--debug', action='store_true', help=argparse.SUPPRESS,
    )

    args = parser.parse_args()
    destdir = getcwd()
    resume = False

    if args.version:
        print(f"{parser.prog} {__version__}")
        sys.exit()

    # Set up logging.
    log_level = logging.WARNING
    if args.verbose:
        log_level = logging.INFO
    if args.debug:
        log_level = logging.DEBUG
    logging.basicConfig(
        level=log_level,
        format=config.MSG_FMT,
        datefmt=config.DATE_FMT,
    )
    logging.debug(f"{args=}")

    if args.continue_download:
        resume = True
    if args.timeout:
        config.HTTP_TIMEOUT = args.timeout
    if args.output_directory:
        destdir = args.output_directory
    destname = None
    if args.filename:
        destname = args.filename
    headers = {}
    for hstr in args.header:
        k, v = hstr.split(':')
        headers[k.strip()] = v.strip()
    return Download(
        url=args.url,
        destdir=destdir,
        destname=destname,
        request_headers=headers,
        resume=resume,
    ).get()



import unittest

from src.net_dl import download


class TestDownload(unittest.TestCase):
    def test_dns_error(self):
        url = 'https://does.not/exist'
        d = download.Download(url)
        self.assertIsNone(d.url.get_head_response())
        self.assertEqual(d.get(), 1)

    def test_404(self):
        url = 'https://httpbin.org/status/404'
        self.assertEqual(download.Download(url).get(), 1)

    def test_30X(self):
        statuses = [301, 302]
        url2 = 'https://httpbin.org'
        url2_esc = 'https%3A%2F%2Fhttpbin.org'
        for s in statuses:
            url1 = f'https://httpbin.org/redirect-to?url={url2_esc}&status={s}'
            d = download.Download(url1)
            self.assertEqual(d.get(), 0)
            self.assertEqual(d.url.final_url, url2)

    def test_is_file(self):
        urls = {
            'https://httpbin.org/html': False,
            'https://httpbin.org/image/svg': True,
            'https://httpbin.org/json': False,
            'https://httpbin.org/robots.txt': False,
            'https://httpbin.org/xml': False,
        }
        for url, result in urls.items():
            dl = download.Download(url)
            dl.url.get_head_response()
            self.assertIs(
                dl.url.is_file,
                result,
                f"{url}: {result}; {dl.url.is_file}"
            )



import requests
import unittest
from pathlib import Path

import src.net_dl


class TestProps(unittest.TestCase):
    def setUp(self):
        self.p = src.net_dl.props.Props(__file__)

    def test_uri(self):
        print(self.p.path)
        assert self.p.path == __file__


class TestLocalFile(unittest.TestCase):
    def setUp(self):
        self.f = Path('./test.txt')
        self.f_text = "test file text\n"
        self.f.write_text(self.f_text)
        self.p = src.net_dl.props.LocalFile(str(self.f))

    def test_path(self):
        self.assertTrue(isinstance(self.p.path, Path))
        self.assertGreater(len(f"{self.p.path}"), 0)
        self.assertEqual(f"{self.p.path}", f"{self.f}")

    def test_size(self):
        self.assertIsNotNone(self.p.size)
        self.assertEqual(self.p.size, len(self.f_text))

    def test_mtime(self):
        old = self.p.get_mtime()
        self.assertIsNotNone(old)
        self.p.set_mtime('Wed, 21 Oct 2015 07:28:00 GMT')
        self.assertNotEqual(old, self.p.get_mtime())

    def tearDown(self):
        self.f.unlink()


class TestUrl(unittest.TestCase):
    def setUp(self):
        self.url = 'https://ip.me'
        self.p = src.net_dl.props.Url(self.url)
        # NOTE: This will make a connection to the URL and be a bit slow.
        self.p.get_head_response()

    def test_init(self):
        self.assertTrue(
            isinstance(
                self.p.head_response.headers,
                requests.structures.CaseInsensitiveDict
            )
        )
        self.assertEqual(
            self.p.head_response.headers.get('Content-Type'),
            'text/plain; charset=utf-8'
        )
        # TODO: Pick a different link to test MD5?



from requests.exceptions import ConnectionError
from requests.exceptions import Timeout
from socket import timeout


DATE_FMT = '%Y-%m-%d %H:%M:%S'
MSG_FMT = '%(asctime)s %(levelname)s: %(message)s'

HTTP_TIMEOUT = 30
HTTP_ERRORS = (
    Timeout,
    ConnectionError,
    timeout,
    Exception,
)



import logging
import re
import requests
import shutil
import sys
import threading
from os import getcwd
from pathlib import Path
from queue import Queue
from time import sleep
from urllib.parse import unquote

from . import config
from .props import LocalFile
from .props import Url


class Download:
    def __init__(
        self,
        url=None,
        destdir=getcwd(),
        destname=None,
        request_headers=None,
        chunk_size=None,
        progress_queue=None,
        remove_on_error=True,
        resume=None,
        timeout=config.HTTP_TIMEOUT,
        callback=None,
        callback_args=list(),
        callback_kwargs=dict(),
    ):
        self.url = Url(url)
        self.is_file = None
        self.destdir = Path(destdir)
        self.request_headers = dict()
        if request_headers:
            self.request_headers = request_headers
        if not self.request_headers.get('Accept-Encoding'):
            self.request_headers['Accept-Encoding'] = 'identity'
        self.chunk_size = chunk_size
        self.progress_queue = progress_queue
        self.remove_on_error = remove_on_error
        self.resume = resume
        self.timeout = timeout
        self.callback = callback
        self.callback_args = callback_args
        self.callback_kwargs = callback_kwargs

    def get(self):
        self.url.get_head_response()
        r = self.url.head_response
        if r is None:
            return 1
        if r.status_code != 200:
            if r.status_code == 404:
                logging.error(f"{r.status_code}: {r.reason}")
                return 1
            else:
                logging.debug(self.url.head_response.__dict__)
                logging.error(f"{r.status_code}: {r.reason}")
                return 1
        if self.chunk_size is None:
            # Ensure at least 2 chunks so that progress queue doesn't hang.
            self.chunk_size = min(round(self.url.size / 2), 100*1024)
        if self.url.is_file:
            self.get_file()
        else:
            self.get_text()
        return 0

    def get_content(self):
        return self._get_completed_request_obj()._content

    def get_text(self):
        ''' Content-Type is "text", return decoded text. '''
        r = self._get_completed_request_obj()
        if r:
            print(r.text)

    def get_file(self, file_mode='wb'):
        ''' Content-Type is "application", show progress and save to file. '''

        # Determine progress queue.
        if self.progress_queue is None:
            use_own_queue = True
            self.progress_queue = Queue()
        else:
            use_own_queue = False

        # Determine output filename.
        cd_str = self.url.head_response.headers.get('Content-Disposition')
        if cd_str:
            filename = self._get_content_disposition_filename(cd_str)
        else:  # get from URL
            filename = unquote(self.url.final_url.split('/')[-1])
        self.dest = LocalFile(self.destdir / filename)
        logging.debug(f"{str(self.dest)=}")

        # If file exists try to resume download.
        self.remaining_size = self.url.size
        if self.dest.path.is_file():
            logging.debug(f"Destination file exists: {self.dest.path}")
            local_size = self.dest.get_size()
            logging.debug(f"Current downloaded size [B]: {local_size}")
            if (
                self.resume
                and local_size < self.remaining_size
                and self.url.head_response.headers.get('Accept-Ranges') == 'bytes'  # noqa: E501
            ):
                logging.debug("Resuming download.")
                file_mode = 'ab'
                if isinstance(self.url.size, int):
                    self.remaining_size = self.url.size - local_size
                    self.request_headers['Range'] = f'bytes={local_size}-{self.url.size}'  # noqa: E501
                else:
                    self.requeset_headers['Range'] = f'bytes={local_size}-'
            elif local_size == self.url.size:
                logging.debug("File already downloaded. Verifying integrity.")
                sum_type = None
                if self.url.md5:
                    sum_type = 'md5'
                if self._check_integrity(sum_type=sum_type):
                    logging.info(f"File already exists: {self.dest.path}")
                    return
                else:
                    logging.debug("Redownloading file.")
            else:
                logging.debug("Local file size mismatch; restarting download.")

        # Log download type.
        if 'Range' in self.request_headers.keys():
            message = f"Continuing download from: {self.url.path}."
        else:
            message = f"Starting new download from: {self.url.path}."
        logging.info(message)

        # Check for available disk space.
        if not self._check_disk_space():
            logging.critical("Not enough disk space.")
            sys.exit(1)

        # Start download thread.
        t = threading.Thread(
            target=self._get_stream_request,
            kwargs={'file_mode': file_mode},
            daemon=True,
        )
        logging.debug("Starting stream request download thread.")
        t.start()
        # Show download progress.
        while t.is_alive():
            if use_own_queue:
                p = self.progress_queue.get()
                self._write_progress_bar(p)
                if p == 100:
                    break
            else:
                sleep(0.1)

        if use_own_queue and sys.stdout.isatty():
            print()  # newline after progress bar is done

        if not self._check_integrity():
            logging.critical("Integrity check failed. Please try again.")
            sys.exit(1)
        logging.info(f"File saved as: {self.dest.path}")

    def _check_disk_space(self):
        free = shutil.disk_usage(self.dest.path.parent).free
        logging.info(f"{self.remaining_size} B needed; {free} B available")
        if self.remaining_size > free:
            return False
        return True

    def _check_integrity(
        self,
        sum_type=None,
    ):
        result = True
        if not self.dest.path.is_file():
            logging.error(f"File does not exist: {self.dest.path}")
            result = False
        if result and self.url.size:
            result = self.dest.size == self.url.size
            logging.debug(f"Same size: {result}")
        if result and sum_type == 'md5':
            result = self.dest.md5 == self.url.md5
            logging.debug(f"Same MD5: {result}")
        return result

    def _get_completed_request_obj(self):
        try:
            r = requests.get(
                str(self.url),
                headers=self.request_headers,
                timeout=self.timeout,
                allow_redirects=True,
            )
        except config.HTTP_ERRORS as e:
            if isinstance(e, requests.exceptions.ConnectionError):
                logging.error(e)
            else:
                logging.error(f"{type(e)}: {e}")
                return
        return r

    def _get_content_disposition_filename(self, cd_header_str):
        """
        'Content-Disposition':
        'attachment;
        filename="=?UTF-8?Q?avast=5Ffree=5Fantivirus=5Foffline=5Fsetup.exe?=";
        filename*=UTF-8\'\'avast_free_antivirus_offline_setup.exe'
        """
        fname = re.findall(
            r"filename\*=([^;]+)",
            cd_header_str,
            flags=re.IGNORECASE
        )
        if not fname:
            fname = re.findall(
                r"filename=([^;]+)",
                cd_header_str,
                flags=re.IGNORECASE
            )
        if fname[0].lower().startswith("utf-8''"):
            fname = re.sub(r"utf-8''", '', fname[0], flags=re.IGNORECASE)
            fname = unquote(fname)
        else:
            fname = fname[0]

        # Remove space and double quotes.
        return fname.strip().strip('"')

    def _get_stream_request(self, file_mode='wb'):
        try:
            with requests.get(
                str(self.url),
                stream=True,
                headers=self.request_headers,
                timeout=self.timeout,
                allow_redirects=True,
            ) as r:
                with self.dest.path.open(mode=file_mode) as f:
                    if file_mode == 'wb':
                        verb = 'Writing'
                    else:
                        verb = 'Appending'
                    logging.debug(f"{verb} data to file: {self.dest.path}")
                    for chunk in r.iter_content(chunk_size=self.chunk_size):
                        f.write(chunk)
                        # Send progress value to queue param.
                        local_size = self.dest.get_size()
                        percent = round(local_size / self.remaining_size * 100)
                        self.progress_queue.put(percent)
                        if self.callback:
                            self.callback(
                                *self.callback_args,
                                **self.callback_kwargs,
                            )
        except config.HTTP_ERRORS as e:
            if sys.stdout.isatty():
                print()
            if isinstance(e, requests.exceptions.ConnectionError):
                logging.error(e)
            else:
                logging.error(f"{type(e)}: {e}")
            if self.remove_on_error:
                logging.info(f"Deleting file: {self.dest.path}")
                self.dest.path.unlink()
            return

        # Set file's mtime from server.
        url_mtime = r.headers.get('Last-Modified')
        if url_mtime:
            self.dest.set_mtime(url_mtime)

    def _write_progress_bar(self, percent):
        screen_width = shutil.get_terminal_size((80, 24)).columns
        y = '.'
        n = ' '
        l_f = int(screen_width * 0.75)  # progress bar length
        l_y = int(l_f * percent / 100)  # num. of chars. complete
        l_n = l_f - l_y  # num. of chars. incomplete
        # end='\x1b[1K\r' to erase to end of line
        if sys.stdout.isatty():
            print(f" [{y * l_y}{n * l_n}] {percent:>3}%", end='\r')
